r"""Utility functions.

"""

import numpy as onp
import jax.numpy as np
from typing import List


def C(n: int) -> np.ndarray:
    r"""The combinatorial matrix :math:`\mathbf C` defined in the paper's
    appendix.

    Args:
        n: the number of sampled haplotypes :math:`n`

    Returns:
        :math:`(n-1)\times(n-1)` matrix

    """
    W1 = np.zeros((n - 1, n - 1))
    W2 = np.zeros((n - 1, n - 1))
    b = np.arange(1, n - 1 + 1)
    # j = 2
    W1 = W1.at[:, 0].set(6 / (n + 1))
    W2 = W2.at[:, 0].set(0)
    # j = 3
    W1 = W1.at[:, 1].set(10 * (5 * n - 6 * b - 4) / (n + 2) / (n + 1))
    W2 = W2.at[:, 1].set((20 * (n - 2)) / (n + 1) / (n + 2))
    for col in range(n - 3):
        # this cast is crucial for floating point precision
        j = np.float64(col + 2)
        # procedurally generated by Zeilberger's algorithm in Mathematica
        W1 = W1.at[:, col + 2].set(
            -(
                (
                    -(
                        (-1 + j)
                        * (1 + j) ** 2
                        * (3 + 2 * j)
                        * (j - n)
                        * (
                            4
                            + 2 * j
                            - 2 * b * j
                            + j**2
                            - b * j**2
                            + 4 * n
                            + 2 * j * n
                            + j**2 * n
                        )
                        * W1[:, col]
                    )
                    - (-1 + 2 * j)
                    * (3 + 2 * j)
                    * (
                        -4 * j
                        - 12 * b * j
                        - 4 * b**2 * j
                        - 6 * j**2
                        - 12 * b * j**2
                        - 2 * b**2 * j**2
                        - 4 * j**3
                        + 4 * b**2 * j**3
                        - 2 * j**4
                        + 2 * b**2 * j**4
                        + 4 * n
                        + 2 * j * n
                        - 6 * b * j * n
                        + j**2 * n
                        - 9 * b * j**2 * n
                        - 2 * j**3 * n
                        - 6 * b * j**3 * n
                        - j**4 * n
                        - 3 * b * j**4 * n
                        + 4 * n**2
                        + 6 * j * n**2
                        + 7 * j**2 * n**2
                        + 2 * j**3 * n**2
                        + j**4 * n**2
                    )
                    * W1[:, col + 1]
                )
                / (
                    j**2
                    * (2 + j)
                    * (-1 + 2 * j)
                    * (1 + j + n)
                    * (3 + b + j**2 - b * j**2 + 3 * n + j**2 * n)
                )
            )
        )  # noqa: E501
        W2 = W2.at[:, col + 2].set(
            (
                (-1 + j)
                * (1 + j)
                * (2 + j)
                * (3 + 2 * j)
                * (j - n)
                * (1 + j - n)
                * (1 + j + n)
                * W2[:, col]
                + (-1 + 2 * j)
                * (3 + 2 * j)
                * (1 + j - n)
                * (j + n)
                * (
                    2
                    - j
                    - 2 * b * j
                    - j**2
                    - 2 * b * j**2
                    + 2 * n
                    + j * n
                    + j**2 * n
                )
                * W2[:, col + 1]
            )
            / ((-1 + j) * j * (2 + j) * (-1 + 2 * j) * (j - n) * (j + n) * (1 + j + n))
        )  # noqa: E501

    return W1 - W2


def M(n: int, t: np.ndarray, y: np.ndarray) -> np.ndarray:
    r"""The matrix :math:`\mathbf M` defined in the paper's appendix

    Args:
        n: the number of sampled haplotypes :math:`n`
        t: time grid, starting at zero and ending at np.inf
        y: population size in each epoch

    Returns:
        :math:`(n-1)\times m` matrix, where :math:`m` is the number of epochs
        (the length of the ``y`` argument)

    """
    # epoch durations
    s = np.diff(t)
    # we handle the final infinite epoch carefully to facilitate autograd
    u = np.exp(-s[:-1] / y[:-1])
    u = np.concatenate((np.array([1]), u, np.array([0])))

    n_range = np.arange(2, n + 1)
    binom_vec = n_range * (n_range - 1) / 2

    return (
        np.exp(
            binom_vec[:, np.newaxis] * np.cumsum(np.log(u[np.newaxis, :-1]), axis=1)
            - np.log(binom_vec[:, np.newaxis])
        )
        @ (np.eye(len(y), k=0) - np.eye(len(y), k=-1))
        @ np.diag(y)
    )


def tmrca_sf(t: np.ndarray, y: np.ndarray, n: int) -> np.ndarray:
    """The survival function of the TMRCA at each time point.

    Args:
        t: time grid (including zero and infinity)
        y: effective population size in each epoch
        n: number of sampled haplotypes
    """
    # epoch durations
    s = np.diff(t)
    logu = -s / y
    logu = np.concatenate((np.array([0]), logu))
    # the A_2j are the product of this matrix
    # NOTE: using letter  "l" as a variable name to match text
    l = onp.arange(2, n + 1)[:, onp.newaxis]  # noqa: E741
    with onp.errstate(divide="ignore"):
        A2_terms = l * (l - 1) / (l * (l - 1) - l.T * (l.T - 1))
    onp.fill_diagonal(A2_terms, 1)
    A2 = np.prod(A2_terms, axis=0)

    binom_vec = l * (l - 1) / 2

    result = np.zeros(len(t))
    result = result.at[:-1].set(
        np.squeeze(
            A2[np.newaxis, :]
            @ np.exp(np.cumsum(logu[np.newaxis, :-1], axis=1)) ** binom_vec
        )
    )

    assert np.all(np.isfinite(result))

    return result


complement = {"A": "T", "C": "G", "G": "C", "T": "A"}


def revcomp(seq: str) -> str:
    """Reverse complement of nucleotide sequence.

    Args:
        seq: nucleotide string (all caps ACGT)
    """
    return "".join(complement[b] for b in reversed(seq))


def misid_partners(mutation_types: List[str]) -> List[int]:
    """Ancestral state misidentification partner indices. Mutation type
    ``None`` is self-partnered.

    Args:
        mutation_types: list of mutation type strings, e.g. [AAA>ACA, ...]

    Examples:

        >>> from mushi import utils

        >>> utils.misid_partners(['TCC>TTC', 'GAA>GGA', None])
        [1, 0, 2]
    """
    to_pair = list(enumerate(mutation_types))
    pair_idxs = [-1] * len(to_pair)
    while to_pair:
        i, mutype1 = to_pair[0]
        if mutype1 is None:
            pair_idxs[i] = i
            del to_pair[0]
        else:
            anc1, der1 = mutype1.split(">")
            match = False
            for new_idx, (j, mutype2) in enumerate(to_pair):
                anc2, der2 = mutype2.split(">")
                if (anc1, der1) in ((der2, anc2), (revcomp(der2), revcomp(anc2))):
                    match = True
                    break
            if not match:
                raise ValueError(
                    "no ancestral misidentification partner found "
                    f"for mutation type {mutype1}"
                )
            pair_idxs[i] = j
            pair_idxs[j] = i
            del to_pair[new_idx]
            if new_idx != 0:
                del to_pair[0]
    assert set(pair_idxs) == set(range(len(mutation_types)))
    return pair_idxs


def mutype_misid(mutation_types: List[str]) -> np.ndarray:
    """mutation type misidentification operator.

    Args:
        mutation_types: list of mutation type strings, e.g. [AAA>ACA, ...]

    Examples:

        >>> from mushi import utils

        >>> utils.mutype_misid(['TCC>TTC', 'GAA>GGA', None])
        DeviceArray([[0., 1., 0.],
                     [1., 0., 0.],
                     [0., 0., 1.]], dtype=float64)
    """
    AM_mut = np.zeros((len(mutation_types), len(mutation_types)))
    for j, i in enumerate(misid_partners(mutation_types)):
        AM_mut = AM_mut.at[i, j].set(1)
    return AM_mut


def fold(x: np.ndarray) -> np.ndarray:
    """transform SFS to folded SFS.

    Args:
        func: loss function name from loss_functions module

    Examples:

        >>> from mushi import utils
        >>> import jax.numpy as np

        >>> sfs = np.array([1000, 100, 10])

        >>> utils.fold(sfs)
        DeviceArray([1010,  100], dtype=int64)
    """
    n = len(x) + 1
    x = (x + x[::-1])[: (n // 2)]
    if n % 2 == 0:
        x = x.at[-1].set(x[-1] // 2)
    return x



r"""
Optimization algorithms.

An abstract base class :class:`mushi.optimization.Optimizer` is defined, from
which concrete optimization classes are derived (perhaps with intermediate
abstraction).

"""

import abc
import numpy as np
from typing import Callable, Tuple
import prox_tv as ptv
from scipy.linalg import cholesky_banded, cho_solve_banded
from functools import lru_cache


class Optimizer(metaclass=abc.ABCMeta):
    """Abstract base class for optimizers.

    Attributes:
        x: solution point

    Args:
        verbose: flag to print convergence messages
    """

    def __init__(self, verbose: bool = False):
        self.verbose = verbose
        self.x = None

    @abc.abstractmethod
    def _initialize(self, x: np.ndarray) -> None:
        """Initialize solution point x, and any auxiliary variables."""
        pass

    def _check_x(self) -> None:
        """Test if x is defined."""
        if self.x is None:
            raise TypeError("solution point x is not initialized")

    @abc.abstractmethod
    def f(self) -> np.float64:
        """Evaluate cost function at current solution point."""
        pass

    @abc.abstractmethod
    def _step(self) -> None:
        """Take an optimization step and update solution point."""
        pass

    def run(
        self, x: np.ndarray, tol: np.float64 = 1e-6, max_iter: int = 100
    ) -> np.ndarray:
        """Optimize until convergence criteria are met.

        Args:
            x: initial point
            tol: relative tolerance in objective function
            max_iter: maximum number of iterations

        Returns:
            x: solution point
        """
        self._initialize(x)
        self._check_x()
        # initial objective value
        f = self.f()
        if self.verbose:
            print(f"initial objective {f:.6e}", flush=True)
        k = 0
        for k in range(1, max_iter + 1):
            self._step()
            if not np.all(np.isfinite(self.x)):
                print("warning: x contains invalid values", flush=True)
            # terminate if objective function is constant within tolerance
            f_old = f
            f = self.f()
            rel_change = np.abs((f - f_old) / f_old)
            if self.verbose:
                print(
                    f"iteration {k}, objective {f:.3e}, "
                    f"relative change {rel_change:.3e}",
                    end="        \r",
                    flush=True,
                )
            if rel_change < tol:
                if self.verbose:
                    print(
                        "\nrelative change in objective function "
                        f"{rel_change:.2g} "
                        f"is within tolerance {tol} after {k} iterations",
                        flush=True,
                    )
                return np.squeeze(self.x)
        if self.verbose and k > 0:
            print(
                f"\nmaximum iteration {max_iter} reached with relative "
                f"change in objective function {rel_change:.2g}",
                flush=True,
            )
        return np.squeeze(self.x)


class LineSearcher(Optimizer):
    """Abstract class for an optimizer with Armijo line search.

    Args:
        s0: initial step size
        max_line_iter: maximum number of line search steps
        gamma: step size shrinkage rate for line search
        verbose: flag to print convergence messages
    """

    def __init__(
        self,
        s0: np.float64 = 1,
        max_line_iter: int = 100,
        gamma: np.float64 = 0.8,
        verbose: bool = False,
    ):
        self.s0 = s0
        self.max_line_iter = max_line_iter
        self.gamma = gamma
        super().__init__(verbose=verbose)


class AccProxGrad(LineSearcher):
    r"""Nesterov accelerated proximal gradient method with backtracking line
    search [1]_.

    The optimization problem solved is:

    .. math::
        \arg\min_x g(x) + h(x)

    where :math:`g` is differentiable, and the proximal operator for :math:`h`
    is available.

    Args:
        g: differentiable term in objective function
        grad: gradient of g
        h: non-differentiable term in objective function
        prox: proximal operator corresponding to h
        verbose: flag to print convergence messages
        line_search_kwargs: line search keyword arguments,
                            see :py:class:`mushi.optimization.LineSearcher`

    References:

        .. [1]
        https://people.eecs.berkeley.edu/~elghaoui/Teaching/EE227A/lecture18.pdf

    Examples:

        >>> import mushi.optimization as opt
        >>> import numpy as np

        We'll use a squared loss term and a Lasso term for this example, since
        we can specify the solution analytically (indeed, it is directly the
        prox of the Lasso term).

        Define :math:`g(x)` and :math:`\boldsymbol\nabla g(x)`:

        >>> def g(x):
        ...     return 0.5 * np.sum(x ** 2)

        >>> def grad(x):
        ...     return x

        Define :math:`h(x)` and :math:`\mathrm{prox}_h(u)` (we will use a Lasso
        term and the corresponding soft thresholding operator):

        >>> def h(x):
        ...     return np.linalg.norm(x, 1)

        >>> def prox(u, s):
        ...     return np.sign(u) * np.clip(np.abs(u) - s, 0, None)

        Initialize optimizer and define initial point

        >>> fista = opt.AccProxGrad(g, grad, h, prox)
        >>> x = np.zeros(2)

        Run optimization

        >>> fista.run(x)
        array([0., 0.])

        Evaluate cost at the solution point

        >>> fista.f()
        0.0

    """

    def __init__(
        self,
        g: Callable[[np.ndarray], np.float64],
        grad: Callable[[np.ndarray], np.float64],
        h: Callable[[np.ndarray], np.float64],
        prox: Callable[[np.ndarray, np.float64], np.float64],
        verbose: bool = False,
        **line_search_kwargs,
    ):
        self.g = g
        self.grad = grad
        self.h = h
        self.prox = prox
        super().__init__(verbose=verbose, **line_search_kwargs)

    def f(self):
        self._check_x()
        return self.g(self.x) + self.h(self.x)

    def _initialize(self, x: np.ndarray) -> None:
        # initialize solution point
        self.x = x
        # initialize momentum iterate
        self.q = self.x
        # initialize step size
        self.s = self.s0
        # initialize step counter
        self.k = 0

    def _step(self) -> None:
        """step with backtracking line search."""
        self._check_x()
        # evaluate differtiable part of objective at momentum point
        g1 = self.g(self.q)
        grad1 = self.grad(self.q)
        if not np.all(np.isfinite(grad1)):
            raise RuntimeError(f"invalid gradient:\n{grad1}")
        # store old iterate
        x_old = self.x
        # Armijo line search
        for line_iter in range(self.max_line_iter):
            # new point via prox-gradient of momentum point
            self.x = self.prox(self.q - self.s * grad1, self.s)
            # G_s(q) as in the notes linked above
            G = (1 / self.s) * (self.q - self.x)
            # test g(q - sG_s(q)) for sufficient decrease
            if self.g(self.q - self.s * G) <= (
                g1 - self.s * (grad1 * G).sum() + (self.s / 2) * (G**2).sum()
            ):
                # Armijo satisfied
                break
            else:
                # Armijo not satisfied
                self.s *= self.gamma  # shrink step size

        # update step count
        self.k += 1
        # update momentum point
        self.q = self.x + ((self.k - 1) / (self.k + 2)) * (self.x - x_old)

        if line_iter == self.max_line_iter - 1:
            print("warning: line search failed", flush=True)
            # reset step size
            self.s = self.s0


class ThreeOpProxGrad(AccProxGrad):
    r"""Three operator splitting proximal gradient method with backtracking
    line search [2]_.

    The optimization problem solved is:

    .. math::
        \arg\min_x g(x) + h_1(x) + h_2(x)

    where :math:`g` is differentiable, and the proximal operators for
    :math:`h_1` and :math:`h_2` are available.

    Args:
        g: differentiable term in objective function
        grad: gradient of g
        h1: 1st non-differentiable term in objective function
        prox1: proximal operator corresponding to h1
        h2: 2nd non-differentiable term in objective function
        prox2: proximal operator corresponding to h2
        verbose: print convergence messages
        line_search_kwargs: line search keyword arguments,
                            see :py:class:`mushi.optimization.LineSearcher`

    References:
        .. [2] Pedregosa, Gidel, Adaptive Three Operator Splitting in
               Proceedings of the 35th International Conference on Machine
               Learning, Proceedings of Machine Learning Research., J. Dy, A.
               Krause, Eds. (PMLR, 2018), pp. 4085–4094.

    Examples:

        Usage is very similar to :meth:`mushi.optimization.AccProxGrad`, except
        that two non-smooth terms (and their associated proximal operators) may
        be specified.

        >>> import mushi.optimization as opt
        >>> import numpy as np

        We'll use a squared loss term, a Lasso term and a box constraint for
        this example.

        Define :math:`g(x)` and :math:`\boldsymbol\nabla g(x)`:

        >>> def g(x):
        ...     return 0.5 * np.sum(x ** 2)

        >>> def grad(x):
        ...     return x

        Define :math:`h_1(x)` and :math:`\mathrm{prox}_{h_1}(u)`.
        We will use a Lasso term and the corresponding soft thresholding
        operator:

        >>> def h1(x):
        ...     return np.linalg.norm(x, 1)

        >>> def prox1(u, s):
        ...     return np.sign(u) * np.clip(np.abs(u) - s, 0, None)

        Define :math:`h_2(x)` and :math:`\mathrm{prox}_{h_2}(u)`. We use a simple
        box constraint on one dimension, although note that this is quite
        artificial, since such constraints don't require operator splitting.

        >>> def h2(x):
        ...     if x[0] < 1:
        ...         return np.inf
        ...     return 0

        >>> def prox2(u, s):
        ...     return np.clip(u, np.array([1, -np.inf]), None)

        Initialize optimizer and define initial point

        >>> threeop = opt.ThreeOpProxGrad(g, grad, h1, prox1, h2, prox2)
        >>> x = np.zeros(2)

        Run optimization

        >>> threeop.run(x)
        array([1., 0.])

        Evaluate cost at the solution point

        >>> threeop.f()
        1.5

    """

    def __init__(
        self,
        g: Callable[[np.ndarray], np.float64],
        grad: Callable[[np.ndarray], np.float64],
        h1: Callable[[np.ndarray], np.float64],
        prox1: Callable[[np.ndarray, np.float64], np.float64],
        h2: Callable[[np.ndarray], np.float64],
        prox2: Callable[[np.ndarray, np.float64], np.float64],
        verbose: bool = False,
        **line_search_kwargs,
    ):
        super().__init__(g, grad, h1, prox1, verbose=verbose, **line_search_kwargs)
        self.h2 = h2
        self.prox2 = prox2

    def f(self):
        return super().f() + self.h2(self.x)

    def _initialize(self, x: np.ndarray) -> None:
        super()._initialize(x)
        # dual variable
        self.u = np.zeros_like(self.q)

    def _step(self) -> None:
        """step with backtracking line search."""
        self._check_x()
        # evaluate differentiable part of objective
        g1 = self.g(self.q)
        grad1 = self.grad(self.q)
        if not np.all(np.isfinite(grad1)):
            raise RuntimeError(f"invalid gradient:\n{grad1}")
        # Armijo line search
        for line_iter in range(self.max_line_iter):
            # new point via prox-gradient of momentum point
            self.x = self.prox(self.q - self.s * (self.u + grad1), self.s)
            # quadratic approximation of objective
            Q = (
                g1
                + (grad1 * (self.x - self.q)).sum()
                + ((self.x - self.q) ** 2).sum() / (2 * self.s)
            )
            if self.g(self.x) - Q <= 0:
                # sufficient decrease satisfied
                break
            else:
                # sufficient decrease not satisfied
                self.s *= self.gamma  # shrink step size
        if line_iter == self.max_line_iter - 1:
            print("warning: line search failed", flush=True)
            # reset step size
            self.s = self.s0

        # update z variables with 2nd prox
        self.q = self.prox2(self.x + self.s * self.u, self.s)
        # update u variables: dual variables
        self.u = self.u + (self.x - self.q) / self.s
        # grow step size
        self.s = min(self.s / self.gamma**2, self.s0)


class TrendFilter(Optimizer):
    r"""Mixed trend filtering via specialized ADMM as in [3]_, section 5.2.

    The optimization problem solved is:

    .. math::
        \arg\min_{\beta \in \mathbb{R}} \frac{1}{2} \|y - \beta\|_2^2
        + \sum_{\ell=1}^r \lambda_\ell \| D^{k_\ell + 1} \beta \|_1

    where :math:`r` is the number of elements of k.

    Args:
        ks: tuple of integer trend filter orders
        lambdas: tuple of penalties corresponding to each k
        rhos: ADMM convergence parameter
        verbose: print convergence messages

    References:
        .. [3] Aaditya Ramdas and Ryan J. Tibshirani.
               Fast and flexible admm algorithms for trend filtering.
               Journal of Computational and Graphical Statistics
               25.3 (2016): 839-858.

    Example:

        >>> import mushi.optimization as opt
        >>> import numpy as np

        Define piecewise constant input signal

        >>> x = np.zeros(10)
        >>> x[3:7] = 1
        >>> x
        array([0., 0., 0., 1., 1., 1., 1., 0., 0., 0.])

        Corrupt the signal with additive noise

        >>> np.random.seed(0)
        >>> x += .1 * np.random.randn(10)

        Initialize and run 0-th order trend filter on input signal

        >>> tf = opt.TrendFilter((0,), (1e-1,))

        >>> tf.run(x)
        array([0.13810345, 0.13809355, 0.13809775, 1.10542391, 1.1054212 ,
               0.99863353, 0.99864752, 0.03853261, 0.03853295, 0.03853668])

    """

    def __init__(
        self,
        ks: Tuple[int],
        lambdas: Tuple[np.float64],
        rhos: Tuple[np.float64] = None,
        verbose: bool = False,
    ):
        self.k = tuple(ks)
        self.λ = tuple(lambdas)
        self.r = len(self.k)
        if rhos is None:
            # default ADMM parameters used in cited paper
            self.ρ = self.λ
        else:
            self.ρ = rhos
        if len(self.λ) != self.r:
            raise ValueError("ks and lambdas must have equal length")
        if len(self.ρ) != self.r:
            raise ValueError("ks and rhos must have equal length")
        super().__init__(verbose=verbose)

    def f(self):
        self._check_x()
        fit = 0.5 * np.sum((self.x - self.y) ** 2)
        penalty = sum(
            λ * np.linalg.norm(np.diff(self.x, k + 1, axis=0), 1)
            for k, λ in zip(self.k, self.λ)
        )
        return fit + penalty

    def _initialize(self, x: np.ndarray) -> None:
        # input signal
        self.y = x
        if self.y.ndim == 1:
            self.y = self.y[:, np.newaxis]
        self.n = self.y.shape[0]
        # initialize solution point (β in the paper cited)
        self.x = np.zeros_like(self.y)
        # initialize Lagrangian dual variables
        self.α = [np.zeros_like(self.y)[: (self.n - k)] for k in self.k]
        self.u = [np.zeros_like(self.y)[: (self.n - k)] for k in self.k]

        self.D, self.DTD = self._D_DTD(self.n, self.k)
        self.c = self._choleskify(self.n, self.k, self.ρ)

    def _step(self):
        self.x = cho_solve_banded(
            (self.c, False),
            self.y
            + sum(
                self.ρ[i] * self.D[i].T @ (self.α[i] + self.u[i]) for i in range(self.r)
            ),
            check_finite=False,
        )
        for i in range(self.r):
            Dx = self.D[i] @ self.x
            for j in range(self.x.shape[1]):
                self.α[i][:, j] = ptv.tv1_1d(
                    Dx[:, j] - self.u[i][:, j], self.λ[i] / self.ρ[i]
                )
            self.u[i] += self.α[i] - Dx

    @staticmethod
    @lru_cache()
    def _D_DTD(n: int, k: Tuple[int]) -> Tuple[Tuple[np.ndarray], Tuple[np.ndarray]]:
        """difference operator for each order in k.

        Args:
            n: number of points in signal
            k: tuple of integer trend filter orders

        Returns:
            D: tuple of difference operator for each k
            DTD: D.T @ D corresponding to above
        """
        D = np.eye(n, k=0) - np.eye(n, k=-1)
        D = tuple(np.linalg.matrix_power(D, k)[k:] for k in k)

        return tuple(zip(*((D[i], D[i].T @ D[i]) for i in range(len(k)))))

    @staticmethod
    @lru_cache(maxsize=1024)
    def _choleskify(n: int, k: Tuple[int], ρ: Tuple[np.float64]) -> np.ndarray:
        """cholesky decomposition needed for linear solves in trend estimate.

        Args:
            n: number of points in signal
            k: tuple of integer trend filter orders
            ρ: ADMM parameter for each order in k

        Returns:
            Cholesky decomposition
        """
        DTD = TrendFilter._D_DTD(n, k)[1]
        A = np.eye(n) + sum(ρ[i] * DTD[i] for i in range(len(k)))
        # A is a banded Hermitian positive definite matrix with upper/lower
        # bandwidth bw. Express in upper diagonal ordered form
        bw = max(k)
        Ab = np.zeros((bw + 1, A.shape[1]))
        for u in range(bw + 1):
            Ab[-(1 + u), u:] = np.diag(A, k=u)

        return cholesky_banded(Ab, check_finite=False)



r"""
Loss functions for measuring goodness of fit.

Each loss function takes an expected data matrix :math:`\mathbb{E}[\mathbf X]`
and an observed data matrix :math:`\mathbf X`, and returns a loss value.
Higher loss means worse fit between :math:`\mathbf X` and
:math:`\mathbb{E}[\mathbf X]`.

Examples
--------

>>> import mushi.loss_functions as lf
>>> import numpy as np

Define expected data matrix :math:`\mathbb{E}[\mathbf X]` and an observed data
matrix :math:`\mathbf X` as :math:`10\times 10` arrays of ones (for this trivial
example):

>>> E = np.ones((10, 10))
>>> X = np.ones((10, 10))

Compute various losses:

- Poisson random field

  >>> lf.prf(E, X)
  DeviceArray(100., dtype=float64)

- Generalized Kullback-Leibler divergence

  >>> lf.dkl(E, X)
  DeviceArray(0., dtype=float64)

- Least squares

  >>> lf.lsq(E, X)
  0.0
"""

import jax.numpy as np


def prf(E: np.ndarray, X: np.ndarray) -> np.float64:
    r"""Poisson random field loss

    .. math::
        \sum_{i,j} \left( \mathbb{E}[X_{i, j}] - X_{i, j} \log \mathbb{E}[X_{i, j}] \right)

    Args:
        E: expectation :math:`\mathbb{E}[\mathbf X]`
        X: data :math:`\mathbf X`

    """
    return (E - X * np.log(E)).sum()


def dkl(E: np.ndarray, X: np.ndarray) -> np.float64:
    r"""Generalized Kullback-Liebler divergence, a Bregman divergence (ignores
    constant term)

    .. math::
        \sum_{i,j} \left(X_{i, j} \log \frac{X_{i, j}}{\mathbb{E}[X_{i, j}]} - X_{i, j} + \mathbb{E}[X_{i, j}] \right)

    Args:
        E: expectation :math:`\mathbb{E}[\mathbf X]`
        X: data :math:`\mathbf X`

    """
    return (X * np.log(X / E) - X + E).sum()


def lsq(E: np.ndarray, X: np.ndarray) -> np.float64:
    r"""Least-squares loss

    .. math::
        \frac{1}{2} \| \mathbb{E}[\mathbf X] - \mathbf X \|_F^2

    Args:
        E: expectation :math:`\mathbb{E}[\mathbf X]`
        X: data :math:`\mathbf X`

    """
    return (1 / 2) * ((E - X) ** 2).sum()



r"""
History objects for representing demography :math:`\eta(t)` and MuSH
:math:`\boldsymbol\mu(t)`.

A base class :class:`mushi.histories.History` is defined, from which
:class:`mushi.eta` and :class:`mushi.mu` are derived and exposed
at package level.

"""

from typing import List

import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns
import mushi.composition as cmp


class History:
    """Base class piecewise constant history. The first epoch starts at zero,
    and the last epoch extends to infinity.

    Args:
        change_points: epoch change points (times)
        vals: constant values for each epoch (rows)
    """

    def __init__(self, change_points: np.array, vals: np.ndarray):
        self.change_points = change_points
        self.vals = vals
        if (
            any(np.diff(self.change_points) <= 0)
            or any(np.isinf(self.change_points))
            or any(self.change_points <= 0)
        ):
            raise ValueError(
                "change_points must be increasing, finite, and " "positive"
            )
        if len(self.vals) != len(self.change_points) + 1:
            raise ValueError(
                f"len(change_points) = {len(self.change_points)}"
                f" implies {len(self.change_points) + 1} epochs,"
                f" but len(vals) = {len(self.vals)}"
            )
        if np.any(self.vals < 0) or np.sum(np.isinf(self.vals)):
            raise ValueError("elements of vals must be finite and nonnegative")
        self.m = len(self.vals)

    def arrays(self):
        """Return time grid and values in each cell."""
        t = np.concatenate((np.array([0]), self.change_points, np.array([np.inf])))
        return t, self.vals

    def epochs(self):
        """Generator yielding epochs as tuples: (start_time, end_time,
        value)"""
        for i in range(self.m):
            if i == 0:
                start_time = 0
            else:
                start_time = self.change_points[i - 1]
            if i == self.m - 1:
                end_time = np.inf
            else:
                end_time = self.change_points[i]
            value = self.vals[i]
            yield (start_time, end_time, value)

    def check_grid(self, other: "History"):
        """Test if time grid is the same as another instance.

        Args:
            other: another History object
        """
        if not np.array_equal(self.change_points, other.change_points):
            raise ValueError("histories have different time grids")

    def plot(self, t_gen: float = None, types=None, **kwargs) -> List[mpl.lines.Line2D]:
        """Plot the history.

        Args:
            t_gen: generation time in years (optional)
            kwargs: key word arguments passed to plt.plot
        """
        t = np.concatenate((np.array([0]), self.change_points))
        if t_gen:
            t *= t_gen
            t_unit = "years ago"
        else:
            t_unit = "generations ago"
        if types is not None:
            idxs = [self.mutation_types.get_loc(type) for type in types]
            vals = self.vals[:, idxs]
        else:
            vals = self.vals
        lines = plt.plot(t, vals, **kwargs)
        plt.xlabel(f"$t$ ({t_unit})")
        plt.xscale("log")
        plt.tight_layout()
        return lines


class eta(History):
    r"""Demographic history :math:`\eta(t)`.

    Args:
        change_points: epoch change points (times)
        y: vector of constant population sizes in each epoch

    Example:

        >>> import mushi
        >>> import numpy as np

        >>> eta = mushi.eta(np.array([1, 10, 100, 1000]),
        ...                 np.array([1000, 100, 100, 1000, 1000]))
    """

    def __init__(self, change_points: np.ndarray, y: np.ndarray):
        super().__init__(change_points, y)
        assert len(self.y.shape) == 1, self.y.shape

    @property
    def y(self):
        r"""read-only alias to ``vals`` attribute in base class."""
        return self.vals

    @y.setter
    def y(self, value):
        self.vals = value

    def plot(self, **kwargs) -> None:
        super().plot(**kwargs)
        plt.ylabel("$\\eta(t)$")
        plt.yscale("log")
        plt.tight_layout()


class mu(History):
    r"""Mutation spectrum history (MuSH) :math:`\boldsymbol\mu(t)`.

    Args:
        change_points: epoch change points (times)
        Z: matrix of constant values for each epoch (rows) in each mutation type
           (columns)
        mutation_types: list of mutation type names (optional ``None`` element)

    Example:

        >>> import mushi
        >>> import numpy as np

        >>> mu = mushi.mu(np.array([1, 10, 100, 1000]),
        ...               np.ones((5, 2)),
        ...               ['TCC>TTC', 'GAA>GGA'])
    """
    mutation_types: List[str]

    def __init__(
        self, change_points: np.ndarray, Z: np.ndarray, mutation_types: List[str]
    ):
        super().__init__(change_points, Z)
        self.mutation_types = mutation_types
        # if mutation rate vector instead of matrix, promote to matrix
        if len(self.Z.shape) == 1:
            self.Z = self.Z[:, np.newaxis]
        assert len(self.Z.shape) == 2, self.Z.shape
        assert len(self.mutation_types) == self.Z.shape[1]
        self.mutation_types = pd.Index(self.mutation_types, name="mutation type")

    @property
    def Z(self):
        """read-only alias to vals attribute in base class."""
        return self.vals

    @Z.setter
    def Z(self, value):
        self.vals = value

    def plot(self, types: List[str] = None, clr: bool = False, **kwargs) -> None:
        """Plot history.

        Args:
            types: list of mutation types to plot (default all)
            clr: flag to normalize to total mutation intensity and display as
                 centered log ratio transform
        """
        lines = super().plot(types=types, **kwargs)
        if clr:
            Z = cmp.clr(self.Z)
            if types is None:
                idxs = range(len(lines))
            else:
                idxs = [self.mutation_types.get_loc(type) for type in types]
            for idx, line in zip(idxs, lines):
                line.set_ydata(Z[:, idx])
            # recompute the ax.dataLim
            ax = plt.gca()
            ax.relim()
            # update ax.viewLim using the new dataLim
            ax.autoscale_view()
            # plt.ylabel('relative mutation intensity')
            plt.ylabel("mutation intensity composition\n(CLR transformed)")
        else:
            plt.ylabel("$\\mu(t)$")
        plt.tight_layout()

    def plot_cumulative(self, t_gen: float = None, clr: bool = False, **kwargs) -> None:
        r"""Plot the cumulative mutation rate, like a Muller plot

        Args:
            t_gen: generation time in years (optional)
            clr: flag to normalize to total mutation intensity and display as
                 centered log ratio transform
            kwargs: key word arguments passed to ``plt.fill_between``
        """
        t = np.concatenate((np.array([0]), self.change_points))
        if t_gen:
            t *= t_gen
            t_unit = "years ago"
        else:
            t_unit = "generations ago"
        Z = np.cumsum(self.Z, axis=1)
        if clr:
            Z = cmp.clr(Z)
        for j in range(Z.shape[1]):
            plt.fill_between(t, Z[:, j - 1] if j else 0, Z[:, j], **kwargs)
        plt.xlabel(f"$t$ ({t_unit})")
        plt.ylabel("$\\mu(t)$")
        plt.xscale("log")
        plt.tight_layout()

    def clustermap(self, t_gen: float = None, **kwargs) -> None:
        r"""Clustermap of compositionally centralized MuSH

        Args:
            t_gen: generation time in years (optional)
            kwargs: additional keyword arguments passed to ``pd.clustermap``
        """
        t = np.concatenate((np.array([0]), self.change_points))
        if t_gen:
            t *= t_gen
            t_unit = "years ago"
        else:
            t_unit = "generations ago"
        Z = cmp.centralize(self.Z)
        label = "mutation intensity\nperturbation"
        df = pd.DataFrame(
            data=Z,
            index=pd.Index(t, name=f"$t$ ({t_unit})"),
            columns=self.mutation_types,
        )
        g = sns.clustermap(df, row_cluster=False, cbar_kws={"label": label}, **kwargs)
        g.ax_heatmap.set_xticklabels(
            g.ax_heatmap.get_xmajorticklabels(), fontsize=9, family="monospace"
        )



import setuptools
import versioneer


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='mushi',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='William DeWitt',
    author_email='wsdewitt@gmail.com',
    description='🍄 [mu]tation [s]pectrum [h]istory [i]nference 🍄',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/harrispopgen/mushi',
    # packages=setuptools.find_packages(exclude=['tests', 'docs']),
    packages=['mushi'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7.*',
    scripts=['bin/mushi'],
    install_requires=[
        'jax',
        'jaxlib',
        'prox-tv-SDIST',
        'seaborn',
        'pandas',
        'matplotlib',
        'numpy',
        'scipy'
    ],
)
